'''
Function:
    Implementation of BuguyyMusicClient: https://buguyy.top/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import copy
from .base import BaseMusicClient
from urllib.parse import urlencode
from rich.progress import Progress
from ..utils import legalizestring, usesearchheaderscookies, resp2json, safeextractfromdict, searchdictbykey, seconds2hms, SongInfo, QuarkParser


'''BuguyyMusicClient'''
class BuguyyMusicClient(BaseMusicClient):
    source = 'BuguyyMusicClient'
    def __init__(self, **kwargs):
        super(BuguyyMusicClient, self).__init__(**kwargs)
        if not self.quark_parser_config.get('cookies'): self.logger_handle.warning(f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so song downloads are restricted and only mp3 files can be downloaded.')
        self.default_search_headers = {
            "accept": "application/json, text/plain, */*",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
            "origin": "https://buguyy.top",
            "priority": "u=1, i",
            "referer": "https://buguyy.top/",
            "sec-ch-ua": "\"Chromium\";v=\"142\", \"Google Chrome\";v=\"142\", \"Not_A Brand\";v=\"99\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-site",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # search rules
        default_rule = {'keyword': keyword}
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'https://a.buguyy.top/newapi/search.php?'
        page_rule = copy.deepcopy(default_rule)
        search_urls = [base_url + urlencode(page_rule)]
        self.search_size_per_page = self.search_size_per_source
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, verify=False, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp=resp)['data']['list']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('id' not in search_result):
                    continue
                song_info = SongInfo(source=self.source)
                # ----parse from quark links
                if self.quark_parser_config.get('cookies'):
                    quark_download_urls = [search_result.get('downurl', ''), search_result.get('ktmdownurl', '')]
                    for quark_download_url in quark_download_urls:
                        song_info = SongInfo(source=self.source)
                        try:
                            m = re.search(r"(?i)(?:WAV|FLAC)#(https?://[^#]+)|MP3#(https?://[^#]+)", quark_download_url)
                            quark_wav_download_url = m.group(1) or m.group(2)
                            download_result, download_url = QuarkParser.parsefromurl(quark_wav_download_url, **self.quark_parser_config)
                            duration = searchdictbykey(download_result, 'duration')
                            duration = [int(float(d)) for d in duration if int(float(d)) > 0]
                            if duration: duration = duration[0]
                            else: duration = 0
                            if not download_url: continue
                            download_url_status = self.quark_audio_link_tester.test(download_url, request_overrides)
                            download_url_status['probe_status'] = self.quark_audio_link_tester.probe(download_url, request_overrides)
                            ext = download_url_status['probe_status']['ext']
                            if ext == 'NULL': ext = 'mp3'
                            song_info.update(dict(
                                download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': download_result},
                                default_download_headers=self.quark_default_download_headers, ext=ext, file_size=download_url_status['probe_status']['file_size'],
                                duration_s=duration, duration=seconds2hms(duration),
                            ))
                            if song_info.with_valid_download_url: break
                        except:
                            continue
                # ----parse from play url
                lyric_result = {}
                if not song_info.with_valid_download_url:
                    song_info = SongInfo(source=self.source)
                    try:
                        resp = self.get(f'https://a.buguyy.top/newapi/geturl2.php?id={search_result["id"]}', verify=False, **request_overrides)
                        resp.raise_for_status()
                        download_result = resp2json(resp=resp)
                        download_url = safeextractfromdict(download_result, ['data', 'url'], '')
                        if not download_url: continue
                        download_url_status = self.audio_link_tester.test(download_url, request_overrides)
                        download_url_status['probe_status'] = self.audio_link_tester.probe(download_url, request_overrides)
                        ext = download_url_status['probe_status']['ext']
                        if ext == 'NULL': download_url.split('?')[0].split('.')[-1] or 'mp3'
                        song_info.update(dict(
                            download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': download_result},
                            ext=ext, file_size=download_url_status['probe_status']['file_size']
                        ))
                    except:
                        continue
                    lyric_result = copy.deepcopy(download_result)
                else:
                    try:
                        resp = self.get(f'https://a.buguyy.top/newapi/geturl2.php?id={search_result["id"]}', verify=False, **request_overrides)
                        resp.raise_for_status()
                        lyric_result = resp2json(resp=resp)
                    except:
                        pass
                if not song_info.with_valid_download_url: continue
                # ----parse more infos
                if not song_info.duration or song_info.duration == '-:-:-':
                    try:
                        duration = '{:02d}:{:02d}:{:02d}'.format(*([0,0,0] + list(map(int, re.findall(r'\d+', safeextractfromdict(lyric_result, ['data', 'duration'], '')))))[-3:])
                        if duration == '00:00:00': duration = '-:-:-'
                    except:
                        duration = '-:-:-'
                else:
                    duration = song_info.duration
                lyric = safeextractfromdict(lyric_result, ['data', 'lrc'], '')
                if not lyric or '歌词获取失败' in lyric: lyric = 'NULL'
                song_info.raw_data['lyric'] = lyric_result
                song_info.update(dict(
                    lyric=lyric, duration=duration, song_name=legalizestring(search_result.get('title', 'NULL'), replace_null_string='NULL'),
                    singers=legalizestring(search_result.get('singer', 'NULL'), replace_null_string='NULL'), 
                    album=legalizestring(safeextractfromdict(lyric_result, ['data', 'album'], ''), replace_null_string='NULL'),
                    identifier=search_result['id'],
                ))
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of FiveSongMusicClient: https://www.5song.xyz/index.html
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
from bs4 import BeautifulSoup
from .base import BaseMusicClient
from rich.progress import Progress
from urllib.parse import urljoin, urlparse
from ..utils import legalizestring, usesearchheaderscookies, searchdictbykey, seconds2hms, SongInfo, QuarkParser


'''settings'''
QUALITY_RANK = {
    "DSD": 0, "WAV": 1, "FLAC": 2, "APE": 3, "ALAC": 4, "AAC": 5, "MP3": 6, "OGG": 7, "M4A": 8,
}


'''FiveSongMusicClient'''
class FiveSongMusicClient(BaseMusicClient):
    source = 'FiveSongMusicClient'
    def __init__(self, **kwargs):
        super(FiveSongMusicClient, self).__init__(**kwargs)
        assert self.quark_parser_config.get('cookies'), f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so the songs cannot be downloaded.'
        self.default_search_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # construct search urls
        self.search_size_per_page = min(self.search_size_per_source, 10)
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            if int(count // page_size) + 1 == 1:
                search_urls.append(f'https://www.5song.xyz/search.html?keyword={keyword}')
            else:
                search_urls.append(f'https://www.5song.xyz/search.html?page={int(count // page_size) + 1}&keyword={keyword}')
            count += page_size
        # return
        return search_urls
    '''_parsesearchresultsfromhtml'''
    def _parsesearchresultsfromhtml(self, html_text: str):
        soup, base_url, search_results = BeautifulSoup(html_text, "html.parser"), "https://www.5song.xyz", []
        for li in soup.select("div.list ul > li"):
            a = li.select_one("a[href]")
            if not a: continue
            href = a.get("href", "").strip()
            detail_url = urljoin(base_url, href)
            title_el = a.select_one("div.con div.t h3")
            title = title_el.get_text(strip=True) if title_el else None
            formats = [s.get_text(strip=True) for s in a.select("div.con div.t span") if s.get_text(strip=True)]
            singer_el = a.select_one("div.singerNum div.singer")
            date_el = a.select_one("div.singerNum div.date")
            num_el = a.select_one("div.singerNum div.num")
            singer = singer_el.get_text(strip=True) if singer_el else None
            date = date_el.get_text(strip=True) if date_el else None
            num = num_el.get_text(strip=True) if num_el else None
            img = a.select_one("div.pic img")
            cover_url = urljoin(base_url, img.get("src")) if img and img.get("src") else None
            search_results.append({"title": title, "formats": formats, "singer": singer, "date": date, "num": num, "detail_url": detail_url, "cover_url": cover_url})
        return search_results
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        base_url = "https://www.5song.xyz"
        def _guessformat(label: str) -> str | None:
            m = re.search(r"(DSD|WAV|FLAC|APE|ALAC|AAC|MP3|OGG|M4A)", label.upper())
            return m.group(1) if m else None
        def _sortbyaudioquality(link_list):
            def _keyfn(x: dict):
                fmt = _guessformat(x.get("label", ""))
                rank = QUALITY_RANK.get(fmt, 999)
                return (rank, fmt or "")
            return sorted(link_list, key=_keyfn)
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = self._parsesearchresultsfromhtml(resp.text)
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('detail_url' not in search_result):
                    continue
                song_info = SongInfo(source=self.source)
                # ----fetch basic information
                try:
                    song_id = urlparse(search_result['detail_url']).path.strip('/').split('/')[-1].split('.')[0]
                    resp = self.get(search_result['detail_url'], **request_overrides)
                    resp.raise_for_status()
                    soup = BeautifulSoup(resp.text, "lxml")
                    quark_links = []
                    for li in soup.select("div.download ul li[data-url]"):
                        quark_url = (li.get("data-url") or "").strip()
                        a = li.select_one("a[href]")
                        label = a.get_text(" ", strip=True) if a else None
                        pc_download_href = a.get("href", "").strip() if a else None
                        pc_download_url = urljoin(base_url, pc_download_href) if pc_download_href else None
                        if "quark" in quark_url: quark_links.append({"label": label, "quark_url": quark_url, "pc_download_url": pc_download_url})
                    download_result = dict(quark_links=quark_links)
                except:
                    continue
                # ----parse from quark links
                sorted_items = _sortbyaudioquality(download_result['quark_links'])
                for item in sorted_items:
                    quark_url = item['quark_url']
                    try:
                        download_result['quark_parse_result'], download_url = QuarkParser.parsefromurl(quark_url, **self.quark_parser_config)
                        duration = searchdictbykey(download_result['quark_parse_result'], 'duration')
                        duration = [int(float(d)) for d in duration if int(float(d)) > 0]
                        if duration: duration = duration[0]
                        else: duration = 0
                        if not download_url: continue
                        download_url_status = self.quark_audio_link_tester.test(download_url, request_overrides)
                        download_url_status['probe_status'] = self.quark_audio_link_tester.probe(download_url, request_overrides)
                        ext = download_url_status['probe_status']['ext']
                        if ext == 'NULL': ext = 'mp3'
                        song_info = SongInfo(
                            source=self.source, download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': download_result},
                            default_download_headers=self.quark_default_download_headers, ext=ext, file_size=download_url_status['probe_status']['file_size'], identifier=song_id, 
                            song_name=legalizestring(search_result.get('title', 'NULL'), replace_null_string='NULL'), album='NULL', duration_s=duration, duration=seconds2hms(duration),
                            singers=legalizestring(search_result.get('singer', 'NULL'), replace_null_string='NULL'),
                        )
                        if song_info.with_valid_download_url: break
                    except:
                        song_info = SongInfo(source=self.source)
                        continue
                if not song_info.with_valid_download_url: continue
                # --lyric results
                try:
                    lyrics_box = soup.select_one("div.viewCon div.text")
                    lyric_result = {'lyrics_box': str(lyrics_box)}
                    lines = [p.get_text(strip=True) for p in lyrics_box.select("p") if p.get_text(strip=True)]
                    lyric = "\n".join(lines)
                except:
                    lyric, lyric_result = 'NULL', {}
                song_info.lyric = lyric
                song_info.raw_data['lyric'] = lyric_result
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of BaseMusicClient
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import copy
import pickle
import requests
from threading import Lock
from rich.text import Text
from itertools import chain
from datetime import datetime
from rich.progress import Task
from fake_useragent import UserAgent
from pathvalidate import sanitize_filepath
from concurrent.futures import ThreadPoolExecutor, as_completed
from rich.progress import (
    Progress, SpinnerColumn, TextColumn, BarColumn, TaskProgressColumn, DownloadColumn, TransferSpeedColumn, TimeRemainingColumn, MofNCompleteColumn, ProgressColumn,
)
from ..utils import (
    LoggerHandle, AudioLinkTester, SongInfo, SongInfoUtils, touchdir, usedownloadheaderscookies, usesearchheaderscookies, cookies2dict, cookies2string, shortenpathsinsonginfos
)


'''AudioAwareColumn'''
class AudioAwareColumn(ProgressColumn):
    def __init__(self):
        super(AudioAwareColumn, self).__init__()
        self._download_col = DownloadColumn()
    '''render'''
    def render(self, task: Task):
        kind = task.fields.get("kind", "download")
        if kind == "overall":
            completed = int(task.completed)
            total = int(task.total) if task.total is not None else 0
            return Text(f"{completed}/{total} audios")
        else:
            return self._download_col.render(task)


'''BaseMusicClient'''
class BaseMusicClient():
    source = 'BaseMusicClient'
    def __init__(self, search_size_per_source: int = 5, auto_set_proxies: bool = False, random_update_ua: bool = False, max_retries: int = 3, maintain_session: bool = False, 
                 logger_handle: LoggerHandle = None, disable_print: bool = False, work_dir: str = 'musicdl_outputs', freeproxy_settings: dict = None, default_search_cookies: dict | str = None,
                 default_download_cookies: dict | str = None, search_size_per_page: int = 10, strict_limit_search_size_per_page: bool = True, quark_parser_config: dict = None):
        # set up work dir
        touchdir(work_dir)
        # set attributes
        self.search_size_per_source = search_size_per_source
        self.auto_set_proxies = auto_set_proxies
        self.random_update_ua = random_update_ua
        self.max_retries = max_retries
        self.maintain_session = maintain_session
        self.logger_handle = logger_handle if logger_handle else LoggerHandle()
        self.disable_print = disable_print
        self.work_dir = work_dir
        self.freeproxy_settings = freeproxy_settings or {}
        self.default_search_cookies = cookies2dict(default_search_cookies)
        self.default_download_cookies = cookies2dict(default_download_cookies)
        self.default_cookies = self.default_search_cookies
        self.search_size_per_page = min(search_size_per_source, search_size_per_page)
        self.strict_limit_search_size_per_page = strict_limit_search_size_per_page
        self.quark_parser_config = quark_parser_config or {}
        # init requests.Session
        self.default_search_headers = {'User-Agent': UserAgent().random}
        self.default_download_headers = {'User-Agent': UserAgent().random}
        self.quark_default_download_headers = {
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.71 Safari/537.36 Core/1.94.225.400 QQBrowser/12.2.5544.400',
            'origin': 'https://pan.quark.cn', 'referer': 'https://pan.quark.cn/', 'accept-language': 'zh-CN,zh;q=0.9', 'cookie': cookies2string(self.quark_parser_config.get('cookies', '')),
        }
        self.quark_default_download_cookies = {} # placeholder, useless now
        self.default_headers = self.default_search_headers
        self._initsession()
        # proxied_session_client
        self.proxied_session_client = None
        if auto_set_proxies:
            from freeproxy import freeproxy
            default_freeproxy_settings = dict(disable_print=True, proxy_sources=['ProxiflyProxiedSession'], max_tries=20, init_proxied_session_cfg={})
            default_freeproxy_settings.update(self.freeproxy_settings)
            self.proxied_session_client = freeproxy.ProxiedSessionClient(**default_freeproxy_settings)
    '''_initsession'''
    def _initsession(self):
        self.session = requests.Session()
        self.session.headers = self.default_headers
        self.audio_link_tester = AudioLinkTester(headers=copy.deepcopy(self.default_download_headers), cookies=copy.deepcopy(self.default_download_cookies))
        self.quark_audio_link_tester = AudioLinkTester(headers=copy.deepcopy(self.quark_default_download_headers), cookies=copy.deepcopy(self.quark_default_download_cookies))
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        raise NotImplementedError('not to be implemented')
    '''_constructuniqueworkdir'''
    def _constructuniqueworkdir(self, keyword: str):
        time_stamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        work_dir = os.path.join(self.work_dir, self.source, f'{time_stamp} {keyword.replace(" ", "")}')
        touchdir(work_dir)
        return work_dir
    '''_removeduplicates'''
    def _removeduplicates(self, song_infos: list[SongInfo] = None) -> list[SongInfo]:
        unique_song_infos, identifiers = [], set()
        for song_info in song_infos:
            if song_info.identifier in identifiers: continue
            identifiers.add(song_info.identifier)
            unique_song_infos.append(song_info)
        return unique_song_infos
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        raise NotImplementedError('not be implemented')
    '''search'''
    @usesearchheaderscookies
    def search(self, keyword: str, num_threadings: int = 5, request_overrides: dict = None, rule: dict = None, main_process_context: Progress = None, main_progress_id: int = None, main_progress_lock: Lock = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # logging
        self.logger_handle.info(f'Start to search music files using {self.source}.', disable_print=self.disable_print)
        # construct search urls
        search_urls = self._constructsearchurls(keyword=keyword, rule=rule, request_overrides=request_overrides)
        # multi threadings for searching music files
        if main_process_context is None:
            owns_progress = True
            main_process_context = Progress(TextColumn("{task.description}"), BarColumn(bar_width=None), MofNCompleteColumn(), TimeRemainingColumn(), refresh_per_second=10)
            main_process_context.__enter__()
        else:
            owns_progress = False
        if main_progress_lock is None: main_progress_lock = Lock()
        with main_progress_lock:
            progress_id = main_process_context.add_task(f"{self.source}.search >>> completed (0/{len(search_urls)})", total=len(search_urls))
            if main_progress_id is not None:
                cur_total = main_process_context.tasks[main_progress_id].total or 0
                main_process_context.update(main_progress_id, total=cur_total + len(search_urls))
                main_process_context.update(main_progress_id, description=f"ALL sources >>> completed ({int(main_process_context.tasks[main_progress_id].completed)}/{cur_total + len(search_urls)})")
        song_infos, submitted_tasks = {}, []
        with ThreadPoolExecutor(max_workers=num_threadings) as pool:
            for search_url_idx, search_url in enumerate(search_urls):
                song_infos[str(search_url_idx)] = []
                submitted_tasks.append(pool.submit(self._search, keyword, search_url, request_overrides, song_infos[str(search_url_idx)], main_process_context, progress_id))
            for future in as_completed(submitted_tasks):
                future.result()
                with main_progress_lock:
                    main_process_context.advance(progress_id, 1)
                    num_searched_urls = int(main_process_context.tasks[progress_id].completed)
                    main_process_context.update(progress_id, description=f"{self.source}.search >>> completed ({num_searched_urls}/{len(search_urls)})")
                    if main_progress_id is None: continue
                    main_process_context.advance(main_progress_id, 1)
                    main_process_context.update(main_progress_id, description=f"ALL sources >>> completed ({int(main_process_context.tasks[main_progress_id].completed)}/{int(main_process_context.tasks[main_progress_id].total or 0)})")
        song_infos = list(chain.from_iterable(song_infos.values()))
        song_infos = self._removeduplicates(song_infos=song_infos)
        work_dir = self._constructuniqueworkdir(keyword=keyword)
        for song_info in song_infos: song_info.work_dir = work_dir
        # logging
        if len(song_infos) > 0:
            work_dir = song_infos[0].work_dir
            touchdir(work_dir)
            self._savetopkl([s.todict() for s in song_infos], os.path.join(work_dir, 'search_results.pkl'))
        else:
            work_dir = self.work_dir
        self.logger_handle.info(f'Finished searching music files using {self.source}. Search results have been saved to {work_dir}, valid items: {len(song_infos)}.', disable_print=self.disable_print)
        if owns_progress: main_process_context.__exit__(None, None, None)
        # return
        return song_infos
    '''_download'''
    @usedownloadheaderscookies
    def _download(self, song_info: SongInfo, request_overrides: dict = None, downloaded_song_infos: list = [], progress: Progress = None, song_progress_id: int = 0):
        request_overrides = request_overrides or {}
        try:
            touchdir(song_info.work_dir)
            if song_info.default_download_headers: request_overrides['headers'] = song_info.default_download_headers
            with self.get(song_info.download_url, stream=True, **request_overrides) as resp:
                resp.raise_for_status()
                total_size, chunk_size, downloaded_size = int(resp.headers.get('content-length', 0)), song_info.get('chunk_size', 1024), 0
                progress.update(song_progress_id, total=total_size)
                with open(song_info.save_path, "wb") as fp:
                    for chunk in resp.iter_content(chunk_size=chunk_size):
                        if not chunk: continue
                        fp.write(chunk)
                        downloaded_size = downloaded_size + len(chunk)
                        if total_size > 0:
                            downloading_text = "%0.2fMB/%0.2fMB" % (downloaded_size / 1024 / 1024, total_size / 1024 / 1024)
                        else:
                            progress.update(song_progress_id, total=downloaded_size)
                            downloading_text = "%0.2fMB/%0.2fMB" % (downloaded_size / 1024 / 1024, downloaded_size / 1024 / 1024)
                        progress.advance(song_progress_id, len(chunk))
                        progress.update(song_progress_id, description=f"{self.source}.download >>> {song_info.song_name} (Downloading: {downloading_text})")
                progress.update(song_progress_id, description=f"{self.source}.download >>> {song_info.song_name} (Success)")
                downloaded_song_infos.append(SongInfoUtils.fillsongtechinfo(copy.deepcopy(song_info), logger_handle=self.logger_handle, disable_print=self.disable_print))
        except Exception as err:
            progress.update(song_progress_id, description=f"{self.source}.download >>> {song_info.song_name} (Error: {err})")
        return downloaded_song_infos
    '''download'''
    @usedownloadheaderscookies
    def download(self, song_infos: list[SongInfo], num_threadings=5, request_overrides: dict = None):
        # init
        request_overrides = request_overrides or {}
        shortenpathsinsonginfos(song_infos=song_infos)
        # logging
        self.logger_handle.info(f'Start to download music files using {self.source}.', disable_print=self.disable_print)
        # multi threadings for downloading music files
        columns = [
            SpinnerColumn(), TextColumn("{task.description}"), BarColumn(bar_width=None), TaskProgressColumn(), AudioAwareColumn(), TransferSpeedColumn(), TimeRemainingColumn(),
        ]
        with Progress(*columns, refresh_per_second=20, expand=True) as progress:
            songs_progress_id = progress.add_task(f"{self.source}.download >>> completed (0/{len(song_infos)})", total=len(song_infos), kind='overall')
            song_progress_ids, downloaded_song_infos, submitted_tasks = [], [], []
            for _, song_info in enumerate(song_infos):
                desc = f"{self.source}.download >>> {song_info.song_name} (Preparing)"
                song_progress_ids.append(progress.add_task(desc, total=None, kind='download'))
            with ThreadPoolExecutor(max_workers=num_threadings) as pool:
                for song_progress_id, song_info in zip(song_progress_ids, song_infos):
                    submitted_tasks.append(pool.submit(self._download, song_info, request_overrides, downloaded_song_infos, progress, song_progress_id))
                for _ in as_completed(submitted_tasks):
                    progress.advance(songs_progress_id, 1)
                    num_downloaded_songs = int(progress.tasks[songs_progress_id].completed)
                    progress.update(songs_progress_id, description=f"{self.source}.download >>> completed ({num_downloaded_songs}/{len(song_infos)})")
        # logging
        if len(downloaded_song_infos) > 0:
            work_dir = downloaded_song_infos[0]['work_dir']
            touchdir(work_dir)
            self._savetopkl([s.todict() for s in downloaded_song_infos], os.path.join(work_dir, 'download_results.pkl'))
        else:
            work_dir = self.work_dir
        self.logger_handle.info(f'Finished downloading music files using {self.source}. Download results have been saved to {work_dir}, valid downloads: {len(downloaded_song_infos)}.', disable_print=self.disable_print)
        # return
        return downloaded_song_infos
    '''get'''
    def get(self, url, **kwargs):
        if 'cookies' not in kwargs: kwargs['cookies'] = self.default_cookies
        resp = None
        for _ in range(self.max_retries):
            if not self.maintain_session:
                self._initsession()
                if self.random_update_ua: self.session.headers.update({'User-Agent': UserAgent().random})
            if self.auto_set_proxies:
                try:
                    self.session.proxies = self.proxied_session_client.getrandomproxy()
                except Exception as err:
                    self.logger_handle.error(f'{self.source}.get >>> {url} (Error: {err})', disable_print=self.disable_print)
                    self.session.proxies = {}
            else:
                self.session.proxies = {}
            proxies = kwargs.pop('proxies', None) or self.session.proxies
            try:
                resp = self.session.get(url, proxies=proxies, **kwargs)
                resp.raise_for_status()
            except Exception as err:
                self.logger_handle.error(f'{self.source}.get >>> {url} (Error: {err})', disable_print=self.disable_print)
                continue
            return resp
        return resp
    '''post'''
    def post(self, url, **kwargs):
        if 'cookies' not in kwargs: kwargs['cookies'] = self.default_cookies
        resp = None
        for _ in range(self.max_retries):
            if not self.maintain_session:
                self._initsession()
                if self.random_update_ua: self.session.headers.update({'User-Agent': UserAgent().random})
            if self.auto_set_proxies:
                try:
                    self.session.proxies = self.proxied_session_client.getrandomproxy()
                except Exception as err:
                    self.logger_handle.error(f'{self.source}.post >>> {url} (Error: {err})', disable_print=self.disable_print)
                    self.session.proxies = {}
            else:
                self.session.proxies = {}
            proxies = kwargs.pop('proxies', None) or self.session.proxies
            try:
                resp = self.session.post(url, proxies=proxies, **kwargs)
                resp.raise_for_status()
            except Exception as err:
                self.logger_handle.error(f'{self.source}.post >>> {url} (Error: {err})', disable_print=self.disable_print)
                continue
            return resp
        return resp
    '''_savetopkl'''
    def _savetopkl(self, data, file_path, auto_sanitize=True):
        if auto_sanitize: file_path = sanitize_filepath(file_path)
        with open(file_path, 'wb') as fp: pickle.dump(data, fp)


'''
Function:
    Implementation of TIDALMusicClient utils
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import sys
import time
import json
import aigpy
import base64
import shutil
import requests
import webbrowser
import subprocess
from .misc import resp2json
from .logger import colorize
from Crypto.Cipher import AES
from mutagen.flac import FLAC
from Crypto.Util import Counter
from urllib.parse import urljoin
from typing import List, Optional, Any
from cryptography.fernet import Fernet
from datetime import datetime, timedelta
from dataclasses import dataclass, field, asdict


'''AV'''
try: import av
except: av = None


'''MediaMetadata'''
class MediaMetadata(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.tags = []


'''StreamUrl'''
class StreamUrl(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.trackid = None
        self.url = None
        self.urls = None
        self.codec = None
        self.encryptionKey = None
        self.soundQuality = None
        self.sampleRate = None
        self.bitDepth = None


'''VideoStreamUrl'''
class VideoStreamUrl(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.codec = None
        self.resolution = None
        self.resolutions = None
        self.m3u8Url = None


'''Artist'''
class Artist(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.id = None
        self.name = None
        self.type = None
        self.picture = None


'''Album'''
class Album(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.id = None
        self.title = None
        self.duration = 0
        self.numberOfTracks = 0
        self.numberOfVideos = 0
        self.numberOfVolumes = 0
        self.releaseDate = None
        self.type = None
        self.version = None
        self.cover = None
        self.videoCover = None
        self.explicit = False
        self.audioQuality = None
        self.audioModes = None
        self.upc = None
        self.popularity = None
        self.copyright = None
        self.streamStartDate = None
        self.mediaMetadata = MediaMetadata()
        self.artist = Artist()
        self.artists = Artist()


'''Playlist'''
class Playlist(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.uuid = None
        self.title = None
        self.numberOfTracks = 0
        self.numberOfVideos = 0
        self.description = None
        self.duration = 0
        self.image = None
        self.squareImage = None


'''Track'''
class Track(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.id = None
        self.title = None
        self.duration = 0
        self.trackNumber = 0
        self.volumeNumber = 0
        self.trackNumberOnPlaylist = 0
        self.version = None
        self.isrc = None
        self.explicit = False
        self.audioQuality = None
        self.audioModes = None
        self.copyRight = None
        self.replayGain = None
        self.peak = None
        self.popularity = None
        self.streamStartDate = None
        self.mediaMetadata = MediaMetadata()
        self.artist = Artist()
        self.artists = Artist()
        self.album = Album()
        self.allowStreaming = False
        self.playlist = None


'''Video'''
class Video(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.id = None
        self.title = None
        self.duration = 0
        self.imageID = None
        self.trackNumber = 0
        self.releaseDate = None
        self.version = None
        self.quality = None
        self.explicit = False
        self.artist = Artist()
        self.artists = Artist()
        self.album = Album()
        self.allowStreaming = False
        self.playlist = None


'''Mix'''
class Mix(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.id = None
        self.tracks = Track()
        self.videos = Video()


'''Lyrics'''
class Lyrics(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.trackId = None
        self.lyricsProvider = None
        self.providerCommontrackId = None
        self.providerLyricsId = None
        self.lyrics = None
        self.subtitles = None


'''SearchDataBase'''
class SearchDataBase(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.limit = 0
        self.offset = 0
        self.totalNumberOfItems = 0


'''SearchAlbums'''
class SearchAlbums(SearchDataBase):
    def __init__(self) -> None:
        super().__init__()
        self.items = Album()


'''SearchArtists'''
class SearchArtists(SearchDataBase):
    def __init__(self) -> None:
        super().__init__()
        self.items = Artist()


'''SearchTracks'''
class SearchTracks(SearchDataBase):
    def __init__(self) -> None:
        super().__init__()
        self.items = Track()


'''SearchVideos'''
class SearchVideos(SearchDataBase):
    def __init__(self) -> None:
        super().__init__()
        self.items = Video()


'''SearchPlaylists'''
class SearchPlaylists(SearchDataBase):
    def __init__(self) -> None:
        super().__init__()
        self.items = Playlist()


'''SearchResult'''
class SearchResult(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.artists = SearchArtists()
        self.albums = SearchAlbums()
        self.tracks = SearchTracks()
        self.videos = SearchVideos()
        self.playlists = SearchPlaylists()


'''StreamRespond'''
class StreamRespond(aigpy.model.ModelBase):
    def __init__(self) -> None:
        super().__init__()
        self.trackid = None
        self.videoid = None
        self.streamType = None
        self.assetPresentation = None
        self.audioMode = None
        self.audioQuality = None
        self.videoQuality = None
        self.manifestMimeType = None
        self.manifest = None


'''SegmentTimelineEntry'''
@dataclass
class SegmentTimelineEntry:
    start_time: Optional[int]
    duration: int
    repeat: int = 0


'''SegmentTemplate'''
@dataclass
class SegmentTemplate:
    media: Optional[str]
    initialization: Optional[str]
    start_number: int = 1
    timescale: int = 1
    presentation_time_offset: int = 0
    timeline: List[SegmentTimelineEntry] = field(default_factory=list)


'''SegmentList'''
@dataclass
class SegmentList:
    initialization: Optional[str]
    media_segments: List[str] = field(default_factory=list)


'''Representation'''
@dataclass
class Representation:
    id: Optional[str]
    bandwidth: Optional[str]
    codec: Optional[str]
    base_url: str
    segment_template: Optional[SegmentTemplate]
    segment_list: Optional[SegmentList]
    '''segments'''
    @property
    def segments(self) -> List[str]:
        if self.segment_list is not None:
            return buildsegmentlist(self.segment_list, self.base_url)
        if self.segment_template is not None:
            return buildsegmenttemplate(self.segment_template, self.base_url, self)
        return []


'''AdaptationSet'''
@dataclass
class AdaptationSet:
    content_type: Optional[str]
    base_url: str
    representations: List[Representation] = field(default_factory=list)


'''Period'''
@dataclass
class Period:
    base_url: str
    adaptation_sets: List[AdaptationSet] = field(default_factory=list)


'''Manifest'''
@dataclass
class Manifest:
    base_url: str
    periods: List[Period] = field(default_factory=list)


'''SessionStorage'''
@dataclass
class SessionStorage:
    access_token: str = None
    refresh_token: str = None
    expires: datetime = None
    user_id: str = None
    country_code: str = None
    device_code: str = None
    user_code: str = None
    '''tojsonbytes'''
    def tojsonbytes(self):
        data = asdict(self)
        if self.expires is not None:
            data["expires"] = self.expires.isoformat()
        else:
            data["expires"] = None
        return json.dumps(data).encode("utf-8")
    '''fromjsonbytes'''
    @classmethod
    def fromjsonbytes(cls, b: bytes):
        data: dict = json.loads(b.decode("utf-8"))
        if data.get("expires"):
            data["expires"] = datetime.fromisoformat(data["expires"])
        else:
            data["expires"] = None
        return cls(**data)
    '''saveencrypted'''
    def saveencrypted(self, path: str, key: bytes = b'3BxQiWxi32p7SCr9SEjGH2Yzj90lxf0EfQ6bi8Vr0dM='):
        f = Fernet(key)
        encrypted = f.encrypt(self.tojsonbytes())
        with open(path, "wb") as fw:
            fw.write(encrypted)
    '''loadencrypted'''
    @classmethod
    def loadencrypted(cls, path: str, key: bytes = b'3BxQiWxi32p7SCr9SEjGH2Yzj90lxf0EfQ6bi8Vr0dM='):
        f = Fernet(key)
        with open(path, "rb") as fr:
            encrypted = fr.read()
        decrypted = f.decrypt(encrypted)
        return cls.fromjsonbytes(decrypted)


'''TIDALTvSession'''
class TIDALTvSession():
    CANDIDATED_CLIENT_ID_SECRETS = [
        {'client_id': '7m7Ap0JC9j1cOM3n', 'client_secret': 'vRAdA108tlvkJpTsGZS8rGZ7xTlbJ0qaZ2K9saEzsgY='},
        {'client_id': '8SEZWa4J1NVC5U5Y', 'client_secret': 'owUYDkxddz+9FpvGX24DlxECNtFEMBxipU0lBfrbq60='},
        {'client_id': 'zU4XHVVkc2tDPo4t', 'client_secret': 'VJKhDFqJPqvsPVNBV6ukXTJmwlvbttP7wlMlrc72se4='},
    ]
    def __init__(self, client_id: str = '7m7Ap0JC9j1cOM3n', client_secret: str = 'vRAdA108tlvkJpTsGZS8rGZ7xTlbJ0qaZ2K9saEzsgY=',
                 headers: dict = None, cookies: dict = None):
        self.session = requests.Session()
        self.client_id = client_id
        self.client_secret = client_secret
        self.headers = {'User-Agent': 'TIDAL_ANDROID/1039 okhttp/3.14.9'}
        self.headers.update(headers or {})
        self.cookies = cookies or {}
        self.storage = SessionStorage()
    '''auth'''
    def auth(self, request_overrides: dict = None):
        # init
        request_overrides = request_overrides or {}
        outputs = dict(
            ok=False, client_id=self.client_id, client_secret=self.client_secret, reason="",
            device_authorization=dict(device_code=None, user_code=None, verification_url=None, auth_check_timeout=None, auth_check_interval=None),
            token=dict(access_token=None, refresh_token=None, expires_in=None), sessions=dict(user_id=None, country_code=None),
        )
        base_url = 'https://auth.tidal.com/v1'
        if 'headers' not in request_overrides: request_overrides['headers'] = self.headers
        if 'cookies' not in request_overrides: request_overrides['cookies'] = self.cookies
        # device authorization
        try:
            resp = self.session.post(f'{base_url}/oauth2/device_authorization', data={'client_id': self.client_id, 'scope': 'r_usr+w_usr+w_sub'}, **request_overrides)
            resp.raise_for_status()
            device_authorization_results = resp2json(resp=resp)
            outputs['device_authorization'] = dict(
                device_code=device_authorization_results['deviceCode'], user_code=device_authorization_results['userCode'], 
                verification_url=device_authorization_results['verificationUri'], auth_check_timeout=device_authorization_results['expiresIn'], 
                auth_check_interval=device_authorization_results['interval']
            )
            self.storage.user_code = device_authorization_results['userCode']
            self.storage.device_code = device_authorization_results['deviceCode']
        except Exception as err:
            outputs['reason'] = f'Device authorization error: {err}'
            return outputs
        # token
        data = {
            'client_id': self.client_id, 'device_code': device_authorization_results['deviceCode'], 'client_secret': self.client_secret,
            'grant_type': 'urn:ietf:params:oauth:grant-type:device_code', 'scope': 'r_usr+w_usr+w_sub',
        }
        user_login_url = 'https://link.tidal.com/' + device_authorization_results['userCode']
        # --if not ssh to server to use musicdl, auto open user_login_url with webbrowser
        is_remote = bool(os.environ.get("SSH_CONNECTION") or os.environ.get("SSH_TTY"))
        if not is_remote:
            try:
                webbrowser.open(user_login_url, new=2)
            except:
                pass
        # --print tips in terminal
        msg = f'Opening {user_login_url} in the browser, log in or sign up to TIDAL manually to continue (in 300 seconds please).'
        print(colorize("TIDAL LOGIN REQUIRED:", 'highlight'))
        print(colorize(msg, 'highlight'))
        # --use tkinter to show tips
        has_display = (
            sys.platform.startswith("win") or sys.platform == "darwin" or bool(os.environ.get("DISPLAY"))
        )
        if has_display:
            import tkinter as tk
            from tkinter import messagebox
            root = tk.Tk()
            root.withdraw()
            root.attributes('-topmost', True)
            messagebox.showinfo("TIDAL Login Required", msg, parent=root)
            root.destroy()
        # --checking user log in or sign up status
        while True:
            resp = self.session.post(f'{base_url}/oauth2/token', data=data, **request_overrides)
            if resp.status_code not in [400]: break
            time.sleep(0.2)
        # --extract required information
        try:
            resp.raise_for_status()
            token_results = resp2json(resp=resp)
            outputs['token'] = dict(
                access_token=token_results['access_token'], refresh_token=token_results['refresh_token'], expires_in=token_results['expires_in']
            )
            self.storage.access_token = token_results['access_token']
            self.storage.refresh_token = token_results['refresh_token']
            self.storage.expires = datetime.now() + timedelta(seconds=token_results['expires_in'])
        except Exception as err:
            outputs['reason'] = f'Token error: {err}'
            return outputs
        # sessions
        request_overrides.pop('headers', {})
        try:
            resp = self.session.get('https://api.tidal.com/v1/sessions', headers=self.auth_headers, **request_overrides)
            resp.raise_for_status()
            sessions_results = resp2json(resp=resp)
            user_id, country_code = sessions_results['userId'], sessions_results['countryCode']
            outputs['sessions'] = dict(user_id=user_id, country_code=country_code)
            self.storage.user_id = user_id
            self.storage.country_code = country_code
        except Exception as err:
            outputs['reason'] = f'Sessions error: {err}'
            return outputs
        # users
        try:
            resp = self.session.get(f'https://api.tidal.com/v1/users/{user_id}?countryCode={country_code}', headers=self.auth_headers, **request_overrides)
            resp.raise_for_status()
        except Exception as err:
            outputs['reason'] = f'Users error: {err}'
            return outputs
        # return
        outputs.update(dict(
            ok=True, reason=f'Successful Routing: {base_url}/oauth2/device_authorization >>> {base_url}/oauth2/token >>> https://api.tidal.com/v1/sessions >>> https://api.tidal.com/v1/users/{user_id}?countryCode={country_code}'
        ))
        return outputs
    '''refresh'''
    def refresh(self, request_overrides: dict = None):
        # init
        request_overrides = request_overrides or {}
        # assert
        assert self.storage.access_token is not None
        # refresh
        base_url = 'https://auth.tidal.com/v1'
        resp = self.session.post(
            f'{base_url}/oauth2/token', 
            data={'refresh_token': self.storage.refresh_token, 'client_id': self.client_id, 'client_secret': self.client_secret, 'grant_type': 'refresh_token'},
            **request_overrides
        )
        resp.raise_for_status()
        token_results = resp2json(resp=resp)
        results = dict(
            refresh_token=token_results.get('refresh_token'), expires=datetime.now()+timedelta(seconds=token_results['expires_in']),
            access_token=token_results['access_token']
        )
        self.storage.access_token = results['access_token']
        self.storage.expires = results['expires']
        self.storage.refresh_token = results['refresh_token'] if results['refresh_token'] else self.storage.refresh_token
        # return
        return results
    '''auth_headers'''
    @property
    def auth_headers(self):
        return {
            'X-Tidal-Token': self.client_id, 'Authorization': 'Bearer {}'.format(self.storage.access_token), 'Connection': 'Keep-Alive',
            'Accept-Encoding': 'gzip', 'User-Agent': 'TIDAL_ANDROID/1039 okhttp/3.14.9'
        }
    '''cache'''
    def cache(self, cache_file_path: str = ''):
        if not cache_file_path:
            cache_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'tidal_tv_session.enc')
        self.storage.saveencrypted(path=cache_file_path)
    '''loadfromcache'''
    def loadfromcache(self, cache_file_path: str = ''):
        if not cache_file_path:
            cache_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'tidal_tv_session.enc')
        if os.path.exists(cache_file_path):
            self.storage = self.storage.loadencrypted(path=cache_file_path)
            return True
        else:
            return False


'''buildsegmentlist'''
def buildsegmentlist(segment_list: SegmentList, base_url: str) -> List[str]:
    segments: List[str] = []
    if segment_list.initialization:
        segments.append(urljoin(base_url, segment_list.initialization))
    for media in segment_list.media_segments:
        segments.append(urljoin(base_url, media))
    return segments


'''completeurl'''
def completeurl(template: str, base_url: str, representation: Representation, *, number: Optional[int] = None, time: Optional[int] = None) -> str:
    mapping = {
        '$RepresentationID$': representation.id, '$Bandwidth$': representation.bandwidth, '$Number$': None if number is None else str(number),
        '$Time$': None if time is None else str(time),
    }
    result = template
    for placeholder, value in mapping.items():
        if value is not None:
            result = result.replace(placeholder, value)
    result = result.replace('$$', '$')
    return urljoin(base_url, result)


'''buildsegmenttemplate'''
def buildsegmenttemplate(template: SegmentTemplate, base_url: str, representation: Representation) -> List[str]:
    segments: List[str] = []
    if template.initialization:
        segments.append(completeurl(template.initialization, base_url, representation))
    number = template.start_number
    current_time: Optional[int] = None
    for entry in template.timeline:
        if entry.start_time is not None:
            current_time = entry.start_time
        elif current_time is None:
            current_time = template.presentation_time_offset
        for _ in range(entry.repeat + 1):
            media = template.media
            if media:
                segments.append(completeurl(media, base_url, representation, number=number, time=current_time))
            number += 1
            if current_time is not None:
                current_time += entry.duration
    return segments


'''decryptsecuritytoken'''
def decryptsecuritytoken(security_token):
    master_key = 'UIlTTEMmmLfGowo/UC60x2H45W6MdGgTRfo/umg4754='
    master_key = base64.b64decode(master_key)
    security_token = base64.b64decode(security_token)
    iv = security_token[:16]
    encrypted_st = security_token[16:]
    decryptor = AES.new(master_key, AES.MODE_CBC, iv)
    decrypted_st = decryptor.decrypt(encrypted_st)
    key = decrypted_st[:16]
    nonce = decrypted_st[16:24]
    return key, nonce


'''decryptfile'''
def decryptfile(efile, dfile, key, nonce):
    counter = Counter.new(64, prefix=nonce, initial_value=0)
    decryptor = AES.new(key, AES.MODE_CTR, counter=counter)
    with open(efile, 'rb') as eflac:
        flac = decryptor.decrypt(eflac.read())
        with open(dfile, 'wb') as dflac:
            dflac.write(flac)


'''ffmpegready'''
def ffmpegready():
    ffmpeg_available = shutil.which("ffmpeg") is not None
    return ffmpeg_available


'''pyavready'''
def pyavready():
    av_available = av is not None
    return av_available


'''remuxwithpyav'''
def remuxwithpyav(src_path: str, dest_path: str):
    if not pyavready(): return False, "PyAV backend unavailable"
    assert av is not None
    try:
        with av.open(src_path) as container:
            audio_stream = next((s for s in container.streams if s.type == "audio"), None)
            if audio_stream is None:
                return False, "PyAV could not locate an audio stream"
            with av.open(dest_path, mode="w", format="flac") as output:
                out_stream = output.add_stream(template=audio_stream)
                for packet in container.demux(audio_stream):
                    if packet.dts is None:
                        continue
                    packet.stream = out_stream
                    output.mux(packet)
    except Exception as exc:
        return False, f"PyAV error: {exc}"
    return os.path.exists(dest_path) and os.path.getsize(dest_path) > 0, "PyAV"


'''remuxwithffmpeg'''
def remuxwithffmpeg(src_path: str, dest_path: str):
    if not ffmpegready():
        return False, "ffmpeg backend unavailable"
    cmd = ["ffmpeg", "-y", "-v", "error", "-i", src_path, "-map", "0:a:0", "-c:a", "copy", dest_path]
    try:
        subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    except subprocess.CalledProcessError as exc:
        return False, f"ffmpeg exited with code {exc.returncode}"
    return os.path.exists(dest_path) and os.path.getsize(dest_path) > 0, "ffmpeg"


'''remuxflacstream'''
def remuxflacstream(src_path: str, dest_path: str):
    if os.path.exists(dest_path): os.remove(dest_path)
    last_reason: Optional[str] = None
    for backend in (remuxwithpyav, remuxwithffmpeg):
        ok, reason = backend(src_path, dest_path)
        if ok: return dest_path, reason
        last_reason = reason
        if os.path.exists(dest_path): os.remove(dest_path)
    return src_path, last_reason


'''formatgain'''
def formatgain(value: Optional[Any]) -> Optional[str]:
    if value is None: return None
    try: return f"{float(value):.2f} dB"
    except (TypeError, ValueError): return str(value)


'''extractmediatags'''
def extractmediatags(track: Track, album: Optional[Album]) -> list[str]:
    tags: list[str] = []
    for source in (getattr(track, "mediaMetadata", None), getattr(album, "mediaMetadata", None) if album else None):
        if source and getattr(source, "tags", None):
            tags = [tag for tag in source.tags if tag]
            if tags: break
    return tags


'''formatpeak'''
def formatpeak(value: Optional[Any]) -> Optional[str]:
    if value is None: return None
    try: return f"{float(value):.6f}"
    except (TypeError, ValueError): return str(value)


'''updateflacmetadata'''
def updateflacmetadata(filepath: str, track: Track, stream: Optional[StreamUrl]):
    # instance
    audio = FLAC(filepath)
    # set tag
    def _settag(key: str, value: Any) -> None:
        if value is None: return
        if isinstance(value, bool):
            text = "1" if value else "0"
            audio[key] = [text]
            return
        if isinstance(value, (list, tuple, set)):
            values = []
            for item in value:
                if item is None: continue
                if isinstance(item, bool): item = "1" if item else "0"
                item_text = str(item).strip()
                if item_text: values.append(item_text)
            if values: audio[key] = values
            return
        text = str(value).strip()
        if text: audio[key] = [text]
    # set tags from track
    _settag("TIDAL_TRACK_ID", track.id)
    _settag("TIDAL_TRACK_VERSION", track.version)
    _settag("TIDAL_TRACK_POPULARITY", track.popularity)
    _settag("TIDAL_STREAM_START_DATE", track.streamStartDate)
    _settag("TIDAL_EXPLICIT", track.explicit)
    _settag("TIDAL_AUDIO_QUALITY", getattr(track, "audioQuality", None))
    _settag("TIDAL_AUDIO_MODES", getattr(track, "audioModes", None) or [])
    _settag("REPLAYGAIN_TRACK_GAIN", formatgain(getattr(track, "replayGain", None)))
    _settag("REPLAYGAIN_TRACK_PEAK", formatpeak(getattr(track, "peak", None)))
    # set tags from stream
    if stream is not None:
        _settag("CODEC", stream.codec)
        _settag("TIDAL_STREAM_SOUND_QUALITY", stream.soundQuality)
        _settag("BITS_PER_SAMPLE", stream.bitDepth)
        _settag("SAMPLERATE", stream.sampleRate)
    # misc
    if track.trackNumberOnPlaylist: _settag("TIDAL_PLAYLIST_TRACK_NUMBER", track.trackNumberOnPlaylist)
    _settag("URL", f"https://listen.tidal.com/track/{track.id}")
    # save
    audio.save()


'''setmetadata'''
def setmetadata(track: Track, filepath: str, stream: Optional[StreamUrl]):
    is_flac_file = filepath.lower().endswith(".flac")
    obj = aigpy.tag.TagTool(filepath)
    obj.album = track.album.title
    obj.title = track.title
    if not aigpy.string.isNull(track.version): obj.title += ' (' + track.version + ')'
    obj.artist = list(map(lambda artist: artist.name, track.artists)) if track.artists else list()
    obj.copyright = track.copyRight
    obj.tracknumber = track.trackNumber
    obj.discnumber = track.volumeNumber
    obj.isrc = track.isrc
    if is_flac_file:
        updateflacmetadata(filepath, track, stream)


'''
Function:
    Implementation of NeteaseMusicClient utils
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import json
import base64
import urllib
import codecs
import urllib.parse
from hashlib import md5
from Crypto.Cipher import AES
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


'''settings'''
MUSIC_QUALITIES = ['jymaster', 'dolby', 'sky', 'jyeffect', 'hires', 'lossless', 'exhigh', 'standard']
DEFAULT_COOKIES = {'MUSIC_U': '1eb9ce22024bb666e99b6743b2222f29ef64a9e88fda0fd5754714b900a5d70d993166e004087dd3b95085f6a85b059f5e9aba41e3f2646e3cebdbec0317df58c119e5'}


'''EapiCryptoUtils'''
class EapiCryptoUtils(object):
    '''hexdigest'''
    @staticmethod
    def hexdigest(data: bytes):
        return "".join([hex(d)[2:].zfill(2) for d in data])
    '''hashdigest'''
    @staticmethod
    def hashdigest(text: str):
        return md5(text.encode("utf-8")).digest()
    '''hashhexdigest'''
    @staticmethod
    def hashhexdigest(text: str):
        return EapiCryptoUtils.hexdigest(EapiCryptoUtils.hashdigest(text))
    '''encryptparams'''
    @staticmethod
    def encryptparams(url: str, payload: dict, aes_key: bytes = b"e82ckenh8dichen8"):
        url_path = urllib.parse.urlparse(url).path.replace("/eapi/", "/api/")
        digest = EapiCryptoUtils.hashhexdigest(f"nobody{url_path}use{json.dumps(payload)}md5forencrypt")
        params = f"{url_path}-36cd479b6b5-{json.dumps(payload)}-36cd479b6b5-{digest}"
        padder = padding.PKCS7(algorithms.AES(aes_key).block_size).padder()
        padded_data = padder.update(params.encode()) + padder.finalize()
        cipher = Cipher(algorithms.AES(aes_key), modes.ECB())
        encryptor = cipher.encryptor()
        enc = encryptor.update(padded_data) + encryptor.finalize()
        return EapiCryptoUtils.hexdigest(enc)


'''WeapiCryptoUtils'''
class WeapiCryptoUtils(object):
    '''createsecretkey'''
    @staticmethod
    def createsecretkey(size: int):
        return (''.join(map(lambda xx: (hex(ord(xx))[2:]), str(os.urandom(size)))))[0: 16]
    '''aesencrypt'''
    @staticmethod
    def aesencrypt(string: str, sec_key: str):
        pad = 16 - len(string) % 16
        if isinstance(string, bytes): string = string.decode('utf-8')
        string = string + str(pad * chr(pad))
        sec_key = sec_key.encode('utf-8')
        encryptor = AES.new(sec_key, 2, b'0102030405060708')
        string = string.encode('utf-8')
        ciphertext = encryptor.encrypt(string)
        ciphertext = base64.b64encode(ciphertext)
        return ciphertext
    '''rsaencrypt'''
    @staticmethod
    def rsaencrypt(string: str, pub_key: str = '010001', modulus: str = '00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7'):
        string = string[::-1]
        rs = int(codecs.encode(string.encode('utf-8'), 'hex_codec'), 16) ** int(pub_key, 16) % int(modulus, 16)
        return format(rs, 'x').zfill(256)
    '''encryptparams'''
    @staticmethod
    def encryptparams(params: dict):
        string = json.dumps(params)
        sec_key = WeapiCryptoUtils.createsecretkey(16)
        enc_string = WeapiCryptoUtils.aesencrypt(string=WeapiCryptoUtils.aesencrypt(string=string, sec_key='0CoJUm6Qyw8W8jud'), sec_key=sec_key)
        enc_sec_key = WeapiCryptoUtils.rsaencrypt(string=sec_key)
        post_data = {'params': enc_string, 'encSecKey': enc_sec_key}
        return post_data